/*
 // Copyright (c) 2017 Intel Corporation
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
 //
 //      http://www.apache.org/licenses/LICENSE-2.0
 //
 // Unless required by applicable law or agreed to in writing, software
 // distributed under the License is distributed on an "AS IS" BASIS,
 // WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 // See the License for the specific language governing permissions and
 // limitations under the License.
 */

#include <iostream>
#include <fstream>
#include <vector>
#include <numeric>
#include <chrono>
#include <string>

#include <gflags/gflags.h>

#include <opencv2/opencv.hpp>
#include <opencv2/dnn.hpp>

#include <ie_plugin_ptr.hpp>
#include <ie_plugin_dispatcher.hpp>
#include <ie_plugin_config.hpp>

#include "network.h"
#include "utils.h"

using namespace cv;
using namespace cv::dnn;

using namespace InferenceEngine;

typedef std::chrono::high_resolution_clock hr_clock;

static const char help_message[] = "Print a usage message.";
static const char model_message[] = "Path to IR .xml file. Required.";
static const char input_message[] = "Path to input image or video. If no input path provided application will use web-cam as video source.";
static const char label_message[] = "Path to label file. Required. ";
static const char mean_message[] = "Mean value. Eg: 127,127,129.";
static const char scale_message[] = "Scale value. Eg: 0.5,0.6,0.7.";
static const char device_message[] = "Device name. CPU by default.";

DEFINE_bool(h, false, help_message);
DEFINE_string(m, "", model_message);
DEFINE_string(i, "", input_message);
DEFINE_string(l, "", label_message);
DEFINE_string(mean, "", mean_message);
DEFINE_string(scale, "", scale_message);
DEFINE_string(d, "CPU", device_message);

typedef struct {
    size_t classId;
    float confidence;
} ClassObjects;

static void ShowUsage()
{
    std::cout << std::endl;
    std::cout << "Options:" << std::endl;
    std::cout << std::endl;
    std::cout << "    -h                        " << help_message << std::endl;

    std::cout << "    -m <path>                 " << model_message << std::endl;
    std::cout << "    -i <path>                 " << input_message << std::endl;
    std::cout << "    -l <path>                 " << label_message << std::endl;

    std::cout << "    -mean <value>             " << mean_message << std::endl;
    std::cout << "    -scale <value>            " << scale_message << std::endl;

    std::cout << "    -d CPU|GPU|FPGA           " << device_message << std::endl;
}


void ParseClassificationResults(InferenceEngine::Blob::Ptr output,
                                size_t topN, std::vector<ClassObjects>& classes) {
    classes.clear();

    SizeVector outDims = output->dims();

    std::vector<unsigned> results;
    std::vector<float> outResults;

    const TBlob<float>::Ptr fOutput = std::dynamic_pointer_cast<TBlob<float>>(output);
    InferenceEngine::TopResults(topN, *fOutput, results);

    for (int i = 0; i < fOutput->size(); i++) {
        outResults.push_back(fOutput->data()[i]);
    }

    size_t batchSize;

    if (outDims.size() == 2) {
        batchSize = outDims[1];
        topN = std::min<unsigned int>((unsigned)outDims[0], topN);
    } else {
        batchSize = outDims[3];
        topN = std::min<unsigned int>((unsigned)outDims[2], topN);
    }

    for (size_t i = 0; i < batchSize; i++) {
        ClassObjects object;

        for (size_t j = 0; j < topN; j++) {
            object.classId = results[i * topN + j];
            object.confidence = static_cast<float>(outResults[object.classId + i * (outResults.size() / batchSize)]);
            
            classes.push_back(object);
        }
    }
}

bool ParseMeanAndScale(size_t inputChannels, std::vector<float>& mean, std::vector<float>& scale) {
    if (ParseFloats(FLAGS_mean, mean, {}) == false) {
        std::cout << "Can't parse mean values. Please use next format: -mean 127.0,127.5,123.4" << std::endl;
        return false;
    }

    if (ParseFloats(FLAGS_scale, scale, {}) == false) {
        std::cout << "Can't parse scale values. Please use next format: -mean 0.07,0.075,0.073" << std::endl;
        return false;
    }

    if (mean.size() == 0) {
        for (size_t i = 0; i < inputChannels; i++) {
            mean.push_back(0.0f);
        }
    }
    else if (mean.size() == 1) {
        for (size_t i = 1; i < inputChannels; i++) {
            mean.push_back(mean[0]);
        }
    }
    else if (mean.size() != inputChannels) {
        std::cout << "Number of mean values should be equal to 1 of to number of network input channels." << std::endl;
        return false;
    }

    if (scale.size() == 0) {
        for (size_t i = 0; i < inputChannels; i++) {
            scale.push_back(1.0f);
        }
    }
    else if (scale.size() == 1) {
        for (size_t i = 1; i < inputChannels; i++) {
            scale.push_back(scale[0]);
        }
    }
    else if (scale.size() != inputChannels) {
        std::cout << "Number of scale values should be equal to 1 of to number of network input channels." << std::endl;
        return false;
    }

    return true;
}

int Infer(const std::string& modelPath, const std::string& imagePath, const std::string& labelsPath) {

	std::shared_ptr<VideoCapture> videoSource;
	auto delay = 33;

	if (modelPath.empty()) {
		std::cout << "A Model IR (.xml) file must be provided." << std::endl;
		ShowUsage();
		return 1;
	}

	if (imagePath.empty()) {
		videoSource = std::make_shared<VideoCapture>(VideoCapture(0));
		delay = 33;
    } else {
    	videoSource = std::make_shared<VideoCapture>(imagePath);
    	delay = 0;
	}

	if (!videoSource->isOpened()) {
		std::cout << "Could not open input stream." << std::endl;
		return 1;
	}

	std::vector<std::string> labels;
	std::string label;

	if (!labelsPath.empty()) {
		ReadLabels(labelsPath.c_str(), labels);
	} else {
		std::cout << "A Classification Label File Must Be Provided." << std::endl;
		ShowUsage();
		return 1;
	}

	std::vector<float> mean;
	std::vector<float> scale;

	int labelsSize = labels.size();

	PluginDispatcher dispatcher({""});
	// TODO: Step 1, Load the Plugin
	InferenceEnginePluginPtr plugin;

	NetworkPtr network = NetworkPtr(new Network(plugin));

	// TODO: Step 2, Read IR Generated by Model Optimizer and
	//       Step 3, Configure Input and Output
	

	std::cout << "Network loaded..." << std::endl;

	std::vector<size_t> inputDims = network->GetInputDims();

	const size_t inputWidth = inputDims[0];
	const size_t inputHeight = inputDims[1];
	const size_t inputChannels = inputDims[2];
	if (ParseMeanAndScale(inputChannels, mean, scale) == false) {
		return 1;
	}

	Mat frame;
	Mat resized;

	std::cout << "Inferencing..." << std::endl << std::endl;

	// TODO: Step 4 Load Model
	

	std::vector<ClassObjects> classes;

	InferenceEngine::Blob::Ptr output = network->GetOutputMap().begin()->second;
	InferenceEngine::Blob::Ptr input = network->GetInputMap().begin()->second;

	float* inputPtr = input->buffer();
	char str[200];

	while (true) {
		*videoSource >> frame;
		if (!frame.data) {
			break;
		}
		resize(frame, resized, Size(inputWidth, inputHeight));
		cvtColor(resized, resized, CV_BGR2RGB);

		// Step 6 Prepare Input
		ConvertBGR(inputWidth, inputHeight, { resized.data }, inputPtr, mean, scale);
		auto start = std::chrono::system_clock::now();

		// TODO: Step 7 Perform Inference
		

		auto end = std::chrono::system_clock::now();
		std::chrono::duration<double> diff = end-start;

		// Step 8 Process Output
		ParseClassificationResults(output, 2, classes); // only top 2

	
		sprintf(str,"Cat/Dog Classification App");
		putText(frame, str,
                Point(5,40), FONT_HERSHEY_DUPLEX, 1, Scalar(255,0,0), 2);

		sprintf(str,"%3.f infers/sec",1.0f/diff.count());
		putText(frame, str,
				Point(5,80), FONT_HERSHEY_DUPLEX, 1, Scalar(255,0,0), 2);
        
		for (size_t i = 0; i < classes.size(); i++) {
			std::cout<< "Class: " << labels[classes[i].classId] << ",  Confidence: " << classes[i].confidence << std::endl;

			putText(frame, labels[classes[i].classId],
					Point(5,i*100+140), FONT_HERSHEY_DUPLEX, 1, Scalar(255,0,0), 2);

			sprintf(str, "%3.1f%% confidence", classes[i].confidence * 100.f);
			putText(frame, str,
					Point(5,i*100+170), FONT_HERSHEY_DUPLEX, 1, Scalar(255,0,0), 2);
		}

		imshow("Inference App", frame);
		auto k = waitKey(delay) & 0xFF;

		if(k == 27)
			break;
	}
	return 0;
}



int main(int argc, char *argv[]) {
	gflags::ParseCommandLineNonHelpFlags(&argc, &argv, true);

	if (FLAGS_h) {
		ShowUsage();
		return 1;
	}

	return Infer(FLAGS_m, FLAGS_i, FLAGS_l);
}
