/*
// Copyright (c) 2017 Intel Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
*/

#include <algorithm>
#include <fstream>
#include <dirent.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <cmath>

#include <x86intrin.h>
#include <string.h>

#include "utils.h"

using namespace InferenceEngine;

static TargetDevice GetDeviceFromStr(const std::string& deviceName) {
    static std::map<std::string, InferenceEngine::TargetDevice> deviceFromNameMap = {
        { "CPU", InferenceEngine::TargetDevice::eCPU },
        { "GPU", InferenceEngine::TargetDevice::eGPU },
        { "FPGA", InferenceEngine::TargetDevice::eFPGA }
    };

    auto val = deviceFromNameMap.find(deviceName);
    return val != deviceFromNameMap.end() ? val->second : InferenceEngine::TargetDevice::eDefault;
}

std::string GetArchPath() {
#ifndef OS_LIB_FOLDER
    #define OS_LIB_FOLDER "/"
#endif
    std::string archPath = "../../../lib/" OS_LIB_FOLDER "ia32";

    if (sizeof(void*) == 8) {
        archPath = "../../../lib/" OS_LIB_FOLDER "intel64";
    }

    return archPath;
}

InferenceEnginePluginPtr SelectPlugin(const std::vector<std::string> &pluginDirs,
                                      const std::string &plugin, TargetDevice device) {
    InferenceEngine::PluginDispatcher dispatcher(pluginDirs);

    if (!plugin.empty()) {
        return dispatcher.getPluginByName(plugin);
    } 
    else {
        return dispatcher.getSuitablePlugin(device);
    }
}

InferenceEnginePluginPtr SelectPlugin(const std::vector<std::string> &pluginDirs,
                                      const std::string &plugin, const std::string& device) {
    return SelectPlugin(pluginDirs, plugin, GetDeviceFromStr(device));
}

char FilePathSeparator() {
#ifdef _WIN32
    return '\\';
#else
    return '/';
#endif
}

std::string FileNameNoExt(const std::string& filePath) {
    auto pos = filePath.rfind('.');

    if (pos == std::string::npos) {
        return filePath;
    }

    return filePath.substr(0, pos);
}

std::string FileNameNoPath(const std::string& filePath) {
    auto pos = filePath.rfind(FilePathSeparator());

    if (pos == std::string::npos) {
        return filePath;
    }

    return filePath.substr(pos + 1);
}

std::string FilePath(const std::string& filePath) {
    auto pos = filePath.rfind(FilePathSeparator());

    if (pos == std::string::npos) {
        return filePath;
    }

    return filePath.substr(0, pos + 1);
}

std::string FileExt(const std::string& filePath) {
    int dot = (int)filePath.find_last_of('.');

    if (dot == -1) {
        return "";
    }
    
    return filePath.substr(dot);
}

bool IsDirectory(const std::string& path) {
    struct stat statbuf;

    if (stat(path.c_str(), &statbuf) != 0) {
        return false;
    }

    return S_ISDIR(statbuf.st_mode);
}

std::string ToUpper(const std::string& str) {
    std::string up;

    std::transform(str.begin(), str.end(), std::back_inserter(up), toupper);

    return up;
}

void GetFilesInDir(const std::string& path, const std::vector<std::string>& exts, std::vector<std::string>& files) {
    if (path.empty()) {
        return;
    }

    std::string fullPath;
    std::string fileName;
    std::string parentPath = path;
    std::string entName;

    DIR *dir;

    struct dirent *ent;

    if ((dir = opendir(path.c_str())) == NULL) {
        return;
    }

    char sep = path[path.size() - 1];

    if (sep != FilePathSeparator()) {
        parentPath += FilePathSeparator();
    }

    bool foundExt;
    std::string fileExt;

    while ((ent = readdir(dir)) != NULL) {
        if (ent->d_name[0] == '.') {
            continue;
        }
        
        if (ent->d_type == DT_DIR) {
            fullPath = parentPath + ent->d_name;

            GetFilesInDir(fullPath, exts, files);
            continue;
        }

        fileName = ent->d_name;

        int pos = fileName.find_last_of(".");
        if (pos == -1 || pos > (fileName.size() - 2)) {
            continue;
        }

        foundExt = false;
        fileExt = ToUpper(fileName.substr(pos));

        for (auto ext : exts) {
            if (fileExt == ext) {
                foundExt = true;
                break;
            }
        }

        if (foundExt) {
            fullPath = parentPath + ent->d_name;
            files.push_back(fullPath);
        }
    }

    closedir(dir);
}

std::string& Trim(std::string &s) {
    s.erase(s.begin(), std::find_if(s.begin(), s.end(), std::not1(std::ptr_fun<int, int>(std::isspace))));
    s.erase(std::find_if(s.rbegin(), s.rend(), std::not1(std::ptr_fun<int, int>(std::isspace))).base(), s.end());
    return s;
}

bool ParseFloats(const std::string& line, std::vector<float>& result, const std::vector<float>& def) {
    result.clear();

    std::istringstream stream(line);
    std::string str;

    if (line.empty()) {
        result = def;
        return true;
    }

    while (getline(stream, str, ',')) {
        try {
            result.push_back(std::stof(str));
        }
        catch (...) {
            return false;
        }
    }

    return true;
}

std::string GetOSSpecificLibraryName(const std::string& lib) {
#ifdef _WIN32
    return lib + ".dll";
#elif __APPLE__
    return "lib" + lib + ".dylib";
#else
    return "lib" + lib + ".so";
#endif
}

void ReadLabels(const char *path, std::vector<std::string> &labels) {
    labels.clear();

    std::string strLine;

    std::ifstream inputFile;
    inputFile.open(path, std::ios::in);

    if (!inputFile.is_open()) {
        return;
    }

    while (std::getline(inputFile, strLine)) {
        labels.push_back(Trim(strLine));
    }
}

void ConvertBGR(const size_t imageWidth, const size_t imageHeight, const std::vector<uint8_t*>& srcs, float* dst,
                const std::vector<float>& mean, const std::vector<float>& scale) {
    const size_t channels = 3;
    const size_t channelSize = imageWidth * imageHeight;
    const size_t imageSize = channelSize * channels;

    float* rBuffer = &dst[channelSize * 0];
    float* gBuffer = &dst[channelSize * 1];
    float* bBuffer = &dst[channelSize * 2];

    const size_t rOff = 2;
    const size_t gOff = 1;
    const size_t bOff = 0;

    size_t idx = 0;

    for (size_t mb = 0; mb < srcs.size(); mb++) {
        const uint8_t* srcBuffer = srcs[mb];

        idx = imageSize * mb;

        for (size_t h = 0; h < imageHeight; h++) {
            size_t srcIdx = h * imageWidth * channels;

            for (size_t w = 0; w < imageWidth; w++, idx++, srcIdx += channels) {
                rBuffer[idx] = ((float)srcBuffer[srcIdx + rOff] - mean[0]) / scale[0];
                gBuffer[idx] =((float)srcBuffer[srcIdx + gOff] - mean[1]) / scale[1];
                bBuffer[idx] = ((float)srcBuffer[srcIdx + bOff] - mean[2]) / scale[2];
            }
        }
    }
}

void ConvertGRAY(const size_t imageWidth, const size_t imageHeight, const std::vector<uint8_t*>& srcs, float* dst) {
    const size_t imageSize = imageWidth * imageHeight;

    for (size_t mb = 0, idx = 0; mb < srcs.size(); mb++, idx++) {
        for (size_t i = 0; i < imageSize; i++) {
            dst[idx] = srcs[mb][i];
        }
    }
}
